# ### Parts of the code are borrowed from Iain Murray's cluster_usage.py,

import sys, os, pexpect, threading, time, re, types

# Things to tweak:
#####################################################################

allowed_time = 10   # Kill connections after allowed_time seconds

# see http://www.gatsby.ucl.ac.uk/~iam23/internal/cluster_usage/help/
use_mytop = True
if use_mytop:
    TOP = r'mytop'
    top_user_trunc = 15
else:
    TOP = 'top'
    top_user_trunc = 8

# Code to connect to machines, gather, cache and load data
#####################################################################
    
class MachineCommunicator(threading.Thread):
    def __init__ (self, machine_name):
        threading.Thread.__init__(self)
        self.machine_name = machine_name
        self.data = None
    def run(self):
        cmd = ' '.join(['ssh -o BatchMode=yes -x -q', self.machine_name,
                        'exec /nfs/home1/iam23/bin/cpu_mytop', TOP])
        self.data = pexpect.run(cmd, timeout=allowed_time)

# Code to process raw gathered data and return easy-to-render lists
#####################################################################

class Usage(object):
    def __init__(self, cpu, mem):
        self.cpu = cpu
        self.mem = mem
    def add(self, cpu, mem):
        self.cpu += cpu
        self.mem += mem

def process_top_output(top_lines):
    # If anyone's ~/.toprc configuration screws up the magic numbers then set use_mytop=True
    num_header_lines = 7
    (user_field, cpu_field, mem_field, prog_field) = (1, 8, 9, 11) # zero-based
    total_kB_memory = int(re.sub(r'Mem: *([0-9]*)k .*', r'\1', top_lines[3]))
    user_usages = {}
    for line in top_lines[num_header_lines:]:
        fields = line.split()
        if len(fields) == 0:
            break
        user = fields[user_field]
        cpu = float(fields[cpu_field])
        mem = float(fields[mem_field])/100*total_kB_memory
        if (cpu < 10) and (mem < 1024*50) and (fields[prog_field] in ("top","sshd","mytop")):
            continue # Ignore minor processes that might be due to this script
        if fields[prog_field] in ("dnetc"):
            continue # Ignore root 'dnetc' process
        if user in user_usages:
            user_usages[user].add(cpu, mem)
        else:
            user_usages[user] = Usage(cpu, mem)
    return (total_kB_memory, user_usages)

def mb_format(mem_in_kB):
    """Nice formatting of mem usage in, e.g. 1024000 -> "1,000 MB"."""
    str = '%d' % int(round(mem_in_kB/1024))
    str = re.sub(r'(...)', r'\1,', str[-1::-1])
    return str[-1::-1].strip(',') + ' MB'

def get_machine_freedom(machine_name):
    # connect to the machine and get the output of 'top'
    print machine_name,
    mc = MachineCommunicator(machine_name)
    mc.run()
    if len(mc.data)==0:
        print ' ? connection failed'
        return 0, 0
    
    # parse data
    machine_data_lines = mc.data.splitlines()
    num_cpus = int(machine_data_lines[0])
    total_kB_memory, user_usages = process_top_output(machine_data_lines[1:])

    # compute number of free cpus and memory
    used_mem = 0
    used_cpu = 0
    for user, usage in user_usages.items():
        used_mem += usage.mem
        used_cpu += usage.cpu
    free_mem = total_kB_memory - used_mem
    free_cpu_perc = num_cpus*100 - used_cpu
    free_cpu = int(free_cpu_perc)/90
    print ' %d cpus, %d free (%.1f perc);' % (num_cpus, free_cpu, free_cpu_perc),
    print ' %s total mem, %s free' % (mb_format(total_kB_memory), mb_format(free_mem))

    return free_mem, free_cpu

# Code to run tasks on remote machines
#####################################################################

def get_ssh_command(machine_name, *ssh_options):
    ssh_options = ' '.join(ssh_options)
    return ' '.join(['ssh ', ssh_options,
                     machine_name])

class RemoteTaskExecuter(threading.Thread):
    def __init__(self, machine_name, task_cmd, logfile=None):
        super(RemoteTaskExecuter, self).__init__()
        self.logfile = logfile
        self.machine_name = machine_name
        self.cmd = task_cmd
        
    def run(self):
        # format command string
        ssh_cmd = get_ssh_command(self.machine_name)
        cmd = ' '.join([ssh_cmd, self.cmd])
        print 'Execute', cmd
        print ' logfile', self.logfile.name
        pexpect.run(cmd, logfile=self.logfile, timeout=None)

def compile_task_list(tasks, default_mem = 1e6):
    task_list = []
    args_list = []
    for task_cmd, args, mems in tasks:
        if mems is None:
            mem_list = [default_mem]*len(args)
        else:
            mem_list = mems
            
        for arg_list in args:
            # make sure that the argument list is a list
            if type(arg_list) is not types.ListType:
                arg_list = [arg_list]
            args_list.append(tuple(arg_list))
            # transform argument list to strings
            arg_list = [str(arg) for arg in arg_list]
            cmd = [task_cmd]
            cmd.extend(arg_list)
            cmd = ' '.join(cmd)
            task_list.append(cmd)
    return task_list, args_list, mem_list

class TaskManager(object):
    def __init__(self, machine_names, tasks, logpath, logname, 
                 sleep_time=300, default_mem=1e6):
        """
        This is the class one should use. It collects a list of machine
        names, a list of tasks, path and name for the log files,
        and send the various tasks around until the list is empty.

        Input arguments:
        
        machine_names --- e.g., ['rachel', 'barlow']
        
        tasks --- a list of tuples [(command_name, arguments, memory)]
        
                  command_name is the command to be executed on the remote machine
                  e.g., '/nfs/home2/berkes/src/simulate_brain'
                  
                  arguments is the list of arguments the command is called with
                  e.g., if arguments is [[1,2], [10,20]], kali is going to
                  send the commands
                  /nfs/home2/berkes/src/simulate_brain 1 2
                  /nfs/home2/berkes/src/simulate_brain 10 20
                  to two different processors

                  memory is a list of estimated memory consumption for every set
                  of arguments, in kB
                  e.g. a memory argument equal to [1e6, 2e6] means that the first
                  command on the list should be sent to a machine with at least
                  1 Gb of memory available, and the second to a machine with at
                  least 2 Gb available.

                  If memory is set to None instead, the minimum memory is given
                  by the default_mem argument
                  
        sleep_time --- check if there are available processors
                       every 'sleep_time' seconds
                       
        default_mem -- if you don't specifiy how much memory every task needs, the
                       defauls_memory (in kB) is used instead
        """
        self.machine_names = machine_names
        self.machine_job_dict = {}
        for name in machine_names:
            self.machine_job_dict[name] = []
        
        self.tasks, self.args, self.mem = compile_task_list(tasks, default_mem)
        self.active_jobs = []
        self.sleep_time = sleep_time
        self.logpath = logpath
        self.logname = logname

    def run_next_task(self, machine_name, idx):
        # process and remove first task
        task_cmd = self.tasks[idx]
        args = self.args[idx]
        logfile = file(os.path.join(self.logpath, self.logname % args), 'w')
        job = RemoteTaskExecuter(machine_name, task_cmd, logfile)
        # add job to active jobs list
        self.active_jobs.append(job)
        # and to machine <-> jobs dictionary
        self.machine_job_dict[job.machine_name].append(job)
        job.start()

    def run(self):
        tasks = self.tasks
        active_jobs = self.active_jobs
        args = self.args
        mem = self.mem
        
        start_time = time.time()
        print '\n====== Welcome to kali, %d tasks in the queue' % len(tasks)
        while len(tasks)>0 or len(active_jobs)>0:
            if len(tasks)>0:
                # collect usage data for all machines
                print 'Collect machine information'
                free_mem = {}
                free_cpu = {}
                for name in self.machine_names:
                    free_mem[name], free_cpu[name] = get_machine_freedom(name)

                # for all machines, if there is enough space, launch next task
                print '\n'
                for name in self.machine_names:
                    # see if we can run new tasks
                    i = 0
                    while i<len(tasks):
                        if free_cpu[name]>0 and free_mem[name]>mem[i] \
                               and len(self.machine_job_dict[name])<8:
                            print name, 'free mem', free_mem[name], 'mem', mem[i]
                            self.run_next_task(name, i)
                            # update machines info
                            free_cpu[name] -= 1
                            free_mem[name] -= mem[i]
                            # delete task
                            del tasks[i:i+1]
                            del args[i:i+1]
                            del mem[i:i+1]
                        else:
                            i += 1
                    # decide how many tasks can be run on this machine
                    #ntasks = min(free_cpu[name], int(free_mem[name]/self.min_memory))
                    #for i in range(min(ntasks, len(tasks))):
                    #    self.run_next_task(name)
            
            # wait sleep_time seconds, then try again
            time.sleep(self.sleep_time)
            
            # remove unactive tasks
            print '\n====== Elapsed time %.2f min' % ((time.time() - start_time)/60.0)
            active = 0
            jobs = map(None, active_jobs)
            for job in jobs:
                if not job.isAlive():
                    self.machine_job_dict[job.machine_name].remove(job)
                    active_jobs.remove(job)
                    print '! Finished: Task', job.cmd, 'on machine', job.machine_name
                else:
                    active += 1
                    print '* Active: Task', job.cmd, 'on machine', job.machine_name
            print '---'
            print '* %d active tasks' % active
            print '* %d tasks in the queue\n' % len(tasks)

#print get_machine_freedom('roy')
